@@ -101,10 +101,10 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 	short numberOfElements = 0;
 	int seqlen = 0;
 	sdp_list_t *pSeq = NULL;
-	uint8_t dataType;
+	uint8_t dataType = expectedType;
 	int status = 0;
 	const uint8_t *p;
-	int bufsize;
+	size_t bufsize;
 
 	scanned = sdp_extract_seqtype(buf, len, &seqType, &data_size);
 
@@ -122,22 +122,26 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 		char *pElem = NULL;
 		int localSeqLength = 0;
 
-		if (bufsize < (int) sizeof(uint8_t)) {
+		if (bufsize < sizeof(uint8_t)) {
 			SDPDBG("->Unexpected end of buffer");
-			return -1;
+			goto failed;
 		}
 
-		dataType = *(uint8_t *)p;
+		if (dataType == SDP_TYPE_ANY)
+			expectedType = *p;
+
+		dataType = *p;
+
 		SDPDBG("Data type: 0x%02x", dataType);
 
 		if (expectedType == SDP_TYPE_UUID) {
 			if (dataType != SDP_UUID16 && dataType != SDP_UUID32 && dataType != SDP_UUID128) {
 				SDPDBG("->Unexpected Data type (expected UUID_ANY)");
-				return -1;
+				goto failed;
 			}
 		} else if (expectedType != SDP_TYPE_ANY && dataType != expectedType) {
 			SDPDBG("->Unexpected Data type (expected 0x%02x)", expectedType);
-			return -1;
+			goto failed;
 		}
 
 		switch (dataType) {
@@ -145,9 +149,9 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 			p += sizeof(uint8_t);
 			seqlen += sizeof(uint8_t);
 			bufsize -= sizeof(uint8_t);
-			if (bufsize < (int) sizeof(uint16_t)) {
+			if (bufsize < sizeof(uint16_t)) {
 				SDPDBG("->Unexpected end of buffer");
-				return -1;
+				goto failed;
 			}
 
 			pElem = malloc(sizeof(uint16_t));
@@ -162,7 +166,7 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 			bufsize -= sizeof(uint8_t);
 			if (bufsize < (int)sizeof(uint32_t)) {
 				SDPDBG("->Unexpected end of buffer");
-				return -1;
+				goto failed;
 			}
 
 			pElem = malloc(sizeof(uint32_t));
@@ -178,7 +182,7 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 			status = sdp_uuid_extract(p, bufsize, (uuid_t *) pElem, &localSeqLength);
 			if (status < 0) {
 				free(pElem);
-				return -1;
+				goto failed;
 			}
 			seqlen += localSeqLength;
 			p += localSeqLength;
@@ -195,7 +199,7 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 			if (seqlen == data_size)
 				break;
 			else if (seqlen > data_size || seqlen > len)
-				return -1;
+				goto failed;
 		} else
 			free(pElem);
 	}
@@ -203,6 +207,10 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 	scanned += seqlen;
 	*pDataType = dataType;
 	return scanned;
+
+failed:
+	sdp_list_free(pSeq, free);
+	return -1;
 }
 
 static int sdp_set_cstate_pdu(sdp_buf_t *buf, sdp_cont_state_t *cstate)
@@ -227,28 +235,39 @@ static int sdp_set_cstate_pdu(sdp_buf_t *buf, sdp_cont_state_t *cstate)
 	return length;
 }
 
-static sdp_cont_state_t *sdp_cstate_get(uint8_t *buffer)
+static int sdp_cstate_get(uint8_t *buffer, size_t len,
+						sdp_cont_state_t **cstate)
 {
-	uint8_t *pdata = buffer;
-	uint8_t cStateSize = *(uint8_t *)pdata;
+	uint8_t cStateSize = *buffer;
+
+	SDPDBG("Continuation State size : %d", cStateSize);
+
+	if (cStateSize == 0) {
+		*cstate = NULL;
+		return 0;
+	}
+
+	buffer++;
+	len--;
+
+	if (len < sizeof(sdp_cont_state_t))
+		return -EINVAL;
 
 	/*
 	 * Check if continuation state exists, if yes attempt
 	 * to get response remainder from cache, else send error
 	 */
-	SDPDBG("Continuation State size : %d", cStateSize);
 
-	pdata += sizeof(uint8_t);
-	if (cStateSize != 0) {
-		sdp_cont_state_t *cstate = malloc(sizeof(sdp_cont_state_t));
-		if (!cstate)
-			return NULL;
-		memcpy(cstate, (sdp_cont_state_t *)pdata, sizeof(sdp_cont_state_t));
-		SDPDBG("Cstate TS : 0x%x", cstate->timestamp);
-		SDPDBG("Bytes sent : %d", cstate->cStateValue.maxBytesSent);
-		return cstate;
-	}
-	return NULL;
+	*cstate = malloc(sizeof(sdp_cont_state_t));
+	if (!(*cstate))
+		return -ENOMEM;
+
+	memcpy(*cstate, buffer, sizeof(sdp_cont_state_t));
+
+	SDPDBG("Cstate TS : 0x%x", (*cstate)->timestamp);
+	SDPDBG("Bytes sent : %d", (*cstate)->cStateValue.maxBytesSent);
+
+	return 0;
 }
 
 /*
@@ -307,15 +326,16 @@ static int service_search_req(sdp_req_t *req, sdp_buf_t *buf)
 	uint32_t cStateId = 0;
 	short *pTotalRecordCount, *pCurrentRecordCount;
 	uint8_t *pdata = req->buf + sizeof(sdp_pdu_hdr_t);
+	size_t data_left = req->len - sizeof(sdp_pdu_hdr_t);
 
-	scanned = extract_des(pdata, req->len - sizeof(sdp_pdu_hdr_t),
-					&pattern, &dtd, SDP_TYPE_UUID);
+	scanned = extract_des(pdata, data_left, &pattern, &dtd, SDP_TYPE_UUID);
 
 	if (scanned == -1) {
 		status = SDP_INVALID_SYNTAX;
 		goto done;
 	}
 	pdata += scanned;
+	data_left -= scanned;
 
 	plen = ntohs(((sdp_pdu_hdr_t *)(req->buf))->plen);
 	mlen = scanned + sizeof(uint16_t) + 1;
@@ -325,18 +345,27 @@ static int service_search_req(sdp_req_t *req, sdp_buf_t *buf)
 		goto done;
 	}
 
+	if (data_left < sizeof(uint16_t)) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
+
 	expected = ntohs(bt_get_unaligned((uint16_t *)pdata));
-	
+
 	SDPDBG("Expected count: %d", expected);
 	SDPDBG("Bytes scanned : %d", scanned);
 
 	pdata += sizeof(uint16_t);
+	data_left -= sizeof(uint16_t);
 
 	/*
 	 * Check if continuation state exists, if yes attempt
 	 * to get rsp remainder from cache, else send error
 	 */
-	cstate = sdp_cstate_get(pdata);
+	if (sdp_cstate_get(pdata, data_left, &cstate) < 0) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
 
 	mtu = req->mtu - sizeof(sdp_pdu_hdr_t) - sizeof(uint16_t) - sizeof(uint16_t) - SDP_CONT_STATE_SIZE;
 	actual = MIN(expected, mtu >> 2);
@@ -566,20 +595,42 @@ static int service_attr_req(sdp_req_t *req, sdp_buf_t *buf)
 	unsigned int max_rsp_size;
 	int status = 0, plen, mlen;
 	uint8_t *pdata = req->buf + sizeof(sdp_pdu_hdr_t);
-	uint32_t handle = ntohl(bt_get_unaligned((uint32_t *)pdata));
+	size_t data_left = req->len - sizeof(sdp_pdu_hdr_t);
+	uint32_t handle;
+
+	if (data_left < sizeof(uint32_t)) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
+
+	handle = ntohl(bt_get_unaligned((uint32_t *)pdata));
 
 	pdata += sizeof(uint32_t);
+	data_left -= sizeof(uint32_t);
+
+	if (data_left < sizeof(uint16_t)) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
+
 	max_rsp_size = ntohs(bt_get_unaligned((uint16_t *)pdata));
+
 	pdata += sizeof(uint16_t);
+	data_left -= sizeof(uint16_t);
+
+	if (data_left < sizeof(sdp_pdu_hdr_t)) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
 
 	/* extract the attribute list */
-	scanned = extract_des(pdata, req->len - sizeof(sdp_pdu_hdr_t),
-						&seq, &dtd, SDP_TYPE_ANY);
+	scanned = extract_des(pdata, data_left, &seq, &dtd, SDP_TYPE_ANY);
 	if (scanned == -1) {
 		status = SDP_INVALID_SYNTAX;
 		goto done;
 	}
 	pdata += scanned;
+	data_left -= scanned;
 
 	plen = ntohs(((sdp_pdu_hdr_t *)(req->buf))->plen);
 	mlen = scanned + sizeof(uint32_t) + sizeof(uint16_t) + 1;
@@ -593,7 +644,10 @@ static int service_attr_req(sdp_req_t *req, sdp_buf_t *buf)
 	 * if continuation state exists, attempt
 	 * to get rsp remainder from cache, else send error
 	 */
-	cstate = sdp_cstate_get(pdata);
+	if (sdp_cstate_get(pdata, data_left, &cstate) < 0) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
 
 	SDPDBG("SvcRecHandle : 0x%x", handle);
 	SDPDBG("max_rsp_size : %d", max_rsp_size);
@@ -686,11 +740,12 @@ static int service_search_attr_req(sdp_req_t *req, sdp_buf_t *buf)
 	short cstate_size = 0;
 	uint8_t dtd = 0;
 	sdp_buf_t tmpbuf;
+	size_t data_left = req->len;
 
 	tmpbuf.data = NULL;
 	pdata = req->buf + sizeof(sdp_pdu_hdr_t);
-	scanned = extract_des(pdata, req->len - sizeof(sdp_pdu_hdr_t),
-					&pattern, &dtd, SDP_TYPE_UUID);
+	data_left = req->len - sizeof(sdp_pdu_hdr_t);
+	scanned = extract_des(pdata, data_left, &pattern, &dtd, SDP_TYPE_UUID);
 	if (scanned == -1) {
 		status = SDP_INVALID_SYNTAX;
 		goto done;
@@ -700,19 +755,35 @@ static int service_search_attr_req(sdp_req_t *req, sdp_buf_t *buf)
 	SDPDBG("Bytes scanned: %d", scanned);
 
 	pdata += scanned;
+	data_left -= scanned;
+
+	if (data_left < sizeof(uint16_t)) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
+
 	max = ntohs(bt_get_unaligned((uint16_t *)pdata));
+
 	pdata += sizeof(uint16_t);
+	data_left -= sizeof(uint16_t);
 
 	SDPDBG("Max Attr expected: %d", max);
 
+	if (data_left < sizeof(sdp_pdu_hdr_t)) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
+
 	/* extract the attribute list */
-	scanned = extract_des(pdata, req->len - sizeof(sdp_pdu_hdr_t),
-						&seq, &dtd, SDP_TYPE_ANY);
+	scanned = extract_des(pdata, data_left, &seq, &dtd, SDP_TYPE_ANY);
 	if (scanned == -1) {
 		status = SDP_INVALID_SYNTAX;
 		goto done;
 	}
+
 	pdata += scanned;
+	data_left -= scanned;
+
 	totscanned += scanned + sizeof(uint16_t) + 1;
 
 	plen = ntohs(((sdp_pdu_hdr_t *)(req->buf))->plen);
@@ -725,7 +796,10 @@ static int service_search_attr_req(sdp_req_t *req, sdp_buf_t *buf)
 	 * if continuation state exists attempt
 	 * to get rsp remainder from cache, else send error
 	 */
-	cstate = sdp_cstate_get(pdata);	// continuation information
+	if (sdp_cstate_get(pdata, data_left, &cstate) < 0) {
+		status = SDP_INVALID_SYNTAX;
+		goto done;
+	}
 
 	svcList = sdp_get_record_list();
 
